import numpy as np
import networkx
import pickle
from models.data import Graph_from_file
from utils import load_model, get_model_attribute
from models.graph_rnn.data import Graph_to_Adj_Matrix
from models.dgmg.data import Graph_to_Action
import scipy
import torch
from torch.utils.data import DataLoader
from torch.utils.data._utils.collate import default_collate as collate
from models.graph_rnn.train import evaluate_loss as eval_loss_graph_rnn
from models.dgmg.train import evaluate_loss as eval_loss_dgmg
from models.graphgen.helper import dfscode_to_tensor
from models.graphgen.train import evaluate_loss as eval_loss_graphgen
from models.gcn.helper import mp_sampler
from datasets.preprocess import get_min_dfscode
import os
import random
from model import create_models


class stats:
    def __init__(self, fname, batch_size, device, cr_iteration, vf2=False):

        self.device = device
        self.args = get_model_attribute('saved_args', fname, self.device)
        self.args.device = device
        self.args.batch_size = batch_size
        self.args.max_cr_iteration = cr_iteration
        self.sample_perm = mp_sampler(self.args, vf2)

        # with open(os.path.join(self.args.current_dataset_path + 'map.dict'), 'rb') as f:
        #     self.feature_map = pickle.load(f)
        self.feature_map = get_model_attribute(
            'feature_map', fname, self.device)
        if self.args.note == "DGMG":
            self.processor = Graph_to_Action(self.args, self.feature_map)
        if self.args.note == "GraphRNN":
            self.processor = Graph_to_Adj_Matrix(self.args, self.feature_map, random_bfs=False)
        if self.args.note == 'Graphgen':
            self.processor = dfscode_to_tensor


        self.model, self.gcn = create_models(self.args, self.feature_map, vf2)
        load_model(fname,  self.device, self.model, gcn=self.gcn, optimizer=None, scheduler=None)
        self.gcn.eval() # do not eval gcn!
        for _, net in self.model.items():
            net.eval()

        random.seed(123)

        self.graphs = []
        for name in os.listdir(self.args.current_dataset_path):
            if name.endswith('.dat'):
                self.graphs.append(len(self.graphs))

        random.shuffle(self.graphs)
        self.graphs_train = self.graphs[:int(0.80 * len(self.graphs))]
        self.graphs_validate = self.graphs[int(0.80 * len(self.graphs)):int(0.90 * len(self.graphs))]

        self.len_eval_all = len(self.graphs_validate)
        self.graphs_validate = self.graphs_validate[::50]
        self.len_eval_partial = len(self.graphs_validate)

        dataset_train = Graph_from_file(self.args, self.graphs_train, self.feature_map)
        dataset_validate = Graph_from_file(self.args, self.graphs_validate, self.feature_map)
        self.dataloader_train = DataLoader(
            dataset_train, batch_size=self.args.batch_size, shuffle=False, drop_last=True,
            num_workers=self.args.num_workers, collate_fn=dataset_train.collate_batch)
        self.dataloader_validate = DataLoader(
            dataset_validate, batch_size=self.args.batch_size, shuffle=False, drop_last=True,
            num_workers=self.args.num_workers, collate_fn=dataset_validate.collate_batch)

    def save_state(self, out):
        save_dict = {
            "estimates_b_m": self.estimates_b_m,
            "log_repetitions_b_m": self.log_repetitions_b_m,
            "log_factorials_b": self.log_factorials_b,
            "all_perms": self.all_perms,
            "sample_size": self.sample_size,
            "args": self.args,
            }
        torch.save(save_dict, out)

    def load_state(self, out):
        save_dict = torch.load(out)
        self.estimates_b_m = save_dict["estimates_b_m"]
        self.log_repetitions_b_m = save_dict["log_repetitions_b_m"]
        self.log_factorials_b = save_dict["log_factorials_b"]
        self.all_perms = save_dict["all_perms"]
        self.sample_size = save_dict["sample_size"]

    def uniform(self, graphs, batch_G, sample_size):
        log_repetitions = torch.empty((self.args.batch_size, sample_size), requires_grad=False)  # (N, M)
        ll_q = torch.empty((self.args.batch_size, sample_size), device=self.args.device)
        # adopt uniform sampling
        batch_num_nodes = [graph.number_of_nodes() for graph in batch_G]
        n_node = sum(batch_num_nodes)
        batch_perms = [[] for _ in range(sample_size)]
        batch_nodes = torch.ones(n_node)

        parameterizations = torch.split(batch_nodes, batch_num_nodes)
        for idx_g in range(len(batch_G)):
            # sample permutation
            batch_perm_g, ll_q_g, log_reps = self.sample_perm(graphs[idx_g]['G'], params=parameterizations[idx_g],
                                                         device=self.args.device, M=sample_size,
                                                         nobfs=self.args.nobfs, max_cr_iteration=self.args.max_cr_iteration)
            # batch_perms.append(perms)
            log_repetitions[idx_g].copy_(log_reps)
            ll_q[idx_g].copy_(ll_q_g)

            for idx_m in range(sample_size):
                batch_perms[idx_m].append(batch_perm_g[idx_m])
        return batch_perms, ll_q, log_repetitions

    def get_joint_estimate(self, loader, sample_size):
        with torch.no_grad():
            estimates_b_m = torch.zeros((len(loader)*self.args.batch_size, sample_size))
            log_factorials_b = torch.zeros(len(loader)*self.args.batch_size)
            log_repetitions_b_m = torch.zeros_like(estimates_b_m)
            all_perms = []
            for batch_id, graphs in enumerate(loader):
                print("batch_id: ", batch_id)
                batch_G = [graph['G'] for graph in graphs]
                if self.args.enable_gcn:
                    if self.args.note != 'Graphgen':
                        batch_perms, ll_q, log_repetitions = self.gcn(graphs, sample_size)
                    elif self.args.note == 'Graphgen':
                        # dfs_code_list used for p(g)
                        batch_perms, ll_q, log_repetitions, dfs_code_list = self.gcn(graphs, sample_size)
                else:
                    batch_perms, ll_q, log_repetitions = self.uniform(graphs, batch_G, sample_size)
                log_factorial = torch.tensor([scipy.special.gammaln(len(perm)+1) for perm in batch_perms[0]])
                log_factorials_b[batch_id*self.args.batch_size:(batch_id+1)*self.args.batch_size].copy_(log_factorial)
                log_repetitions_b_m[batch_id*self.args.batch_size:(batch_id+1)*self.args.batch_size].copy_(log_repetitions)
                all_perms.extend(batch_perms)

                if self.args.note == "GraphRNN":
                    for m in range(sample_size):
                        data = [self.processor(graph, perms) for graph, perms in zip(batch_G, batch_perms[m])]
                        data = collate(data)

                        nll_p = eval_loss_graph_rnn(self.args, self.model, data, self.feature_map)

                        ll_p = -nll_p
                        ll_p_q = ll_p-ll_q[:, m]
                        estimates_b_m[batch_id*self.args.batch_size:(batch_id+1)*self.args.batch_size, m].copy_(ll_p_q)

                if self.args.note == "DGMG":
                    for m in range(sample_size):
                        data = [self.processor(graph, perms) for graph, perms in zip(batch_G, batch_perms[m])]
                        # data = processor.collate_batch(data)
                        nll_p = eval_loss_dgmg(self.model['dgmg'], data)
                        # nll_prob = eval_loss_graph_rnn(args, model, data, feature_map)
                        ll_p = -nll_p
                        ll_p_q = ll_p-ll_q[:, m]
                        estimates_b_m[batch_id*self.args.batch_size:(batch_id+1)*self.args.batch_size, m].copy_(ll_p_q)
                if self.args.note == 'Graphgen':
                    for m in range(sample_size):
                        data = [self.processor(code, self.feature_map) for code in dfs_code_list[m::sample_size]]
                        data = collate(data)
                        nll_p = eval_loss_graphgen(self.args, self.model, data, self.feature_map)
                        ll_p = -nll_p
                        ll_p_q = ll_p-ll_q[:, m]
                        estimates_b_m[batch_id * self.args.batch_size:(batch_id + 1) * self.args.batch_size, m].copy_(
                            ll_p_q)


            estimates_b_m = estimates_b_m

            self.estimates_b_m = estimates_b_m
            self.log_repetitions_b_m = log_repetitions_b_m
            self.log_factorials_b = log_factorials_b
            self.all_perms = all_perms
            self.sample_size = sample_size

    def dfs_likelihood(self, loader):
        dfs_codes =[]
        for batch_id, graphs in enumerate(loader):
            print("batch_id: ", batch_id)
            G = [graph['G'] for graph in graphs][0]
            dfs_code = get_min_dfscode(G)
            dfs_codes.append(dfs_code)

        data = [self.processor(code, self.feature_map) for code in dfs_codes]


        nll_p = eval_loss_graphgen(self.args, self.model, data, self.feature_map)
        nll_p = torch.mean(nll_p).item()
        return -nll_p








    def model_likelihood(self,n_samples):
        estimates_b_m, log_repetitions_b_m, log_factorials_b = self.estimates_b_m, self.log_repetitions_b_m, self.log_factorials_b
        # log_factorials_b_m = log_factorials_b.unsqueeze(-1).repeat(1, self.sample_size)
        estimates_b_m = estimates_b_m - log_repetitions_b_m #- log_factorials_b_m
            # print(estimates_b)
        estimates_b = torch.logsumexp(estimates_b_m[:, :n_samples], dim=1)
        estimates_b = estimates_b - torch.log(torch.tensor(n_samples, dtype=estimates_b.dtype))# +log_factorials_b
        estimates = torch.mean(estimates_b, dim=0)
        # print(estimates.item())
        return estimates.item()

    def model_elbo(self):
        estimates_b_m, log_repetitions_b_m, log_factorials_b = self.estimates_b_m, self.log_repetitions_b_m, self.log_factorials_b
        estimates_b_m = estimates_b_m - log_repetitions_b_m
        elbo = torch.mean(estimates_b_m)
        # print(estimates_b)
        # elbo = torch.mean(estimates_b)
        return elbo.item()


if __name__ == '__main__':
    # fname = "output/GraphRNN_caveman_small_gat_nobfs_2020_12_21_18_00_40/model_save/epoch_241.dat"
    fnames = {
        #"ENZYMES_gat_GRNN": "output/Graphgen_ENZYMES_gat_nobfs_2021_01_05_10_27_51/model_save/epoch_299.dat",
        "Lung_gat_GRNN": "output/Graphgen_Yeast_gat_nobfs_2021_01_12_06_54_36/model_save/epoch_1.dat"
        #"citeseer_gat_GRNN": "output/Graphgen_citeseer_small_gat_nobfs_2021_01_18_02_13_14/model_save/epoch_99.dat",


    }

    # fname = "/home/golf/Downloads/epoch_1000.dat"
    # fname = "/media/golf/DEBC83BFBC8390A7/Users/golf/Documents/GitHub/graph_generation/output/GraphRNN_caveman_small_unif_nobfs_2020_12_21_18_03_12/model_save/epoch_500.dat"
    # sta = stats(fname=fname, batch_size=1, device='cuda', cr_iteration=3, vf2=True)
    # sta.get_joint_estimate(sta.dataloader_validate, 1000)
    # # # print(sta.log_repetitions_b_m)
    # # sta.load_state("cavement_vf2_likelihood.pt")
    # # print(torch.mean(sta.log_repetitions_b_m))
    # ll_cr = sta.model_likelihood()
    # elbo = sta.model_elbo()
    # # print(elbo)
    # print(torch.mean(sta.log_repetitions_b_m))
    # print(ll_cr, elbo)
    # tensor(21.5489)
    # -5908.79345703125 - 429.2673645019531
    results = {}
    for setting, fname in fnames.items():
        sta = stats(fname=fname, batch_size=1, device='cuda', cr_iteration=3, vf2=False)
        sta.get_joint_estimate(sta.dataloader_validate, 100)
        sta.save_state("{}.pt".format(setting))
        elbo = sta.model_elbo()
        print('{} elbo'.format(setting))
        print(elbo)
        #ll_crs = []
        # for n_samples in range(50, 1001, 50):
        #     ll_cr = sta.model_likelihood(n_samples)
        #     ll_crs.append(ll_cr)
        ll_cr = sta.model_likelihood(50)

        results[setting] = ll_cr
        print('{} ll_cr'.format(setting))
        print(ll_cr)
    print(results)

        # print(torch.mean(sta.log_repetitions_b_m))
        # print(ll_cr, elbo)
    # sta = stats(fname=fname, batch_size=1, device='cuda', cr_iteration=2, vf2=False)
    # sta.get_joint_estimate(sta.dataloader_train, 6400)
    # # print(sta.log_repetitions_b_m)
    # # sta.load_state("cavement_cr_likelihood_2.pt")
    # ll_cr = sta.model_likelihood()
    # elbo = sta.model_elbo()
    # # print(elbo)
    # print(torch.mean(sta.log_repetitions_b_m))
    # print(ll_cr, elbo)
    #
    #
    # sta = stats(fname=fname, batch_size=1, device='cuda', cr_iteration=1, vf2=False)
    # sta.get_joint_estimate(sta.dataloader_train, 6400)
    # # print(sta.log_repetitions_b_m)
    # # sta.load_state("cavement_cr_likelihood_1.pt")
    # ll_cr = sta.model_likelihood()
    # elbo = sta.model_elbo()
    # # print(elbo)
    # print(torch.mean(sta.log_repetitions_b_m))
    # print(ll_cr, elbo)
    #
    #
    # sta = stats(fname=fname, batch_size=1, device='cuda', cr_iteration=0, vf2=False)
    # sta.get_joint_estimate(sta.dataloader_train, 6400)
    # # print(sta.log_repetitions_b_m)
    # # sta.save_state("cavement_cr_likelihood_0.pt")
    # ll_cr = sta.model_likelihood()
    # elbo = sta.model_elbo()
    # # print(elbo)
    # print(torch.mean(sta.log_repetitions_b_m))
    # print(ll_cr, elbo)
