import torch
import torch.nn as nn
import torch.nn.functional as F

from dgl.nn.pytorch import GATConv

"""
    GAT: Graph Attention Network
    Graph Attention Networks (Veličković et al., ICLR 2018)
    https://arxiv.org/abs/1710.10903
"""


class GATHeadLayer(nn.Module):
    def __init__(self, in_dim, out_dim, dropout, batch_norm):
        super().__init__()
        self.dropout = dropout

        self.batch_norm = batch_norm

        self.fc = nn.Linear(in_dim, out_dim, bias=False)
        self.attn_fc = nn.Linear(2 * out_dim, 1, bias=False)
        self.batchnorm_h = nn.BatchNorm1d(out_dim)

    def edge_attention(self, edges):
        z2 = torch.cat([edges.src['z'], edges.dst['z']], dim=1)
        a = self.attn_fc(z2)
        return {'e': F.leaky_relu(a)}

    def message_func(self, edges):
        return {'z': edges.src['z'], 'e': edges.data['e']}

    def reduce_func(self, nodes):
        alpha = F.softmax(nodes.mailbox['e'], dim=1)
        alpha = F.dropout(alpha, self.dropout, training=self.training)
        h = torch.sum(alpha * nodes.mailbox['z'], dim=1)
        return {'h': h}

    def forward(self, g, h):
        z = self.fc(h)
        g.ndata['z'] = z
        g.apply_edges(self.edge_attention)
        g.update_all(self.message_func, self.reduce_func)
        h = g.ndata['h']

        if self.batch_norm:
            h = self.batchnorm_h(h)
        h = F.elu(h)
        h = F.dropout(h, self.dropout, training=self.training)
        return h

# Delete graph_norm
class GATLayer(nn.Module):
    """
        Param: [in_dim, out_dim, n_heads]
    """

    def __init__(self, in_dim, out_dim, num_heads, dropout,  batch_norm, residual=False, activation=None,
                 dgl_builtin=False):

        super().__init__()
        self.dgl_builtin = dgl_builtin

        if dgl_builtin == False:
            self.in_channels = in_dim
            self.out_channels = out_dim
            self.num_heads = num_heads
            self.residual = residual

            if in_dim != (out_dim * num_heads):
                self.residual = False

            self.heads = nn.ModuleList()
            for i in range(num_heads):
                self.heads.append(GATHeadLayer(in_dim, out_dim, dropout, batch_norm))
            self.merge = 'cat'

        else:
            self.in_channels = in_dim
            self.out_channels = out_dim
            self.num_heads = num_heads
            self.residual = residual
            self.activation = activation

            self.batch_norm = batch_norm

            if in_dim != (out_dim * num_heads):
                self.residual = False

            # Both feat and weighting dropout tied together here
            self.conv = GATConv(in_dim, out_dim, num_heads, dropout, dropout)
            self.batchnorm_h = nn.BatchNorm1d(out_dim)

    def forward(self, g, h):
        if self.dgl_builtin == False:
            h_in = h  # for residual connection
            head_outs = [attn_head(g, h) for attn_head in self.heads]

            if self.merge == 'cat':
                h = torch.cat(head_outs, dim=1)
            else:
                h = torch.mean(torch.stack(head_outs))

            if self.residual:
                h = h_in + h  # residual connection
            return h
        else:
            h_in = h  # for residual connection

            h = self.conv(g, h).flatten(1)


            if self.batch_norm:
                h = self.batchnorm_h(h)

            if self.residual:
                h = h_in + h  # residual connection

            if self.activation:
                h = self.activation(h)
            return h

    def __repr__(self):
        return '{}(in_channels={}, out_channels={}, heads={}, residual={})'.format(self.__class__.__name__,
                                                                                   self.in_channels,
                                                                                   self.out_channels, self.num_heads,
                                                                                   self.residual)

